import os
from abc import abstractmethod

from rooms_shared_services.src.lambda_handlers.abstract.processors import AbstractLambdaDataProcessor
from rooms_shared_services.src.settings.dynamodb.customer_leads import Settings as CustomerLeadStorageSettings
from rooms_shared_services.src.storage.dynamodb import DynamodbStorageClient
from rooms_shared_services.src.storage.models import BaseDynamodbModel


class AbstractCustomerLeadStorageProcessor(AbstractLambdaDataProcessor):
    storage_client_method: str

    def __init__(self) -> None:
        """Set attributes.

        Raises:
            ValueError: When no key id provided
        """
        key_id = os.environ.get("AWS_ACCESS_KEY_ID")
        if not key_id:
            raise ValueError("No key id")
        storage_settings = CustomerLeadStorageSettings()
        storage_client_params = {
            "tablename": storage_settings.tablename,
            "region_name": storage_settings.region_name,
            "endpoint_url": storage_settings.endpoint_url,
        }
        self.storage_client = self.provide_storage_client(**storage_client_params)
        self.model_item: BaseDynamodbModel | None = None

    def provide_storage_client(self, **storage_client_params):
        return DynamodbStorageClient(**storage_client_params)

    @abstractmethod
    def provide_storage_method_params(self, *args, **kwargs) -> dict:
        ...

    def run(self, *args, **kwargs):
        storage_method = getattr(self.storage_client, self.storage_client_method)
        storage_method_params = self.provide_storage_method_params(*args, **kwargs)
        return storage_method(**storage_method_params)



from pydantic_settings import SettingsConfigDict

from rooms_shared_services.src.settings.pricing.base import PriceMarginSettings


class SignalPriceMarginSettings(PriceMarginSettings):
    model_config = SettingsConfigDict(env_prefix="signal_")



from pydantic import Field


class Unset:  # noqa: WPS306
    def __bool__(self) -> bool:
        """Make always false.

        Returns:
            bool: _description_
        """
        return False


UnsetField = Field(default_factory=Unset)



from typing import Literal
from uuid import UUID

from rooms_shared_services.src.storage.models import BaseDynamodbModel

TaskStatus = Literal["scheduled", "pending", "processing", "succeded", "failed"]


class DelayedTaskKey(BaseDynamodbModel):
    task_id: UUID
    attempt_number: int


class DelayedTask(BaseDynamodbModel):
    task_id: UUID
    attempt_number: int
    task_variant: str = ""
    scheduled_at: str | None = None
    pending_at: str | None = None
    processed_at: str | None = None
    succeded_at: str | None = None
    failed_at: str | None = None
    task_status: TaskStatus
    task_content: dict

    @property
    def key(self):
        return DelayedTaskKey(task_id=self.task_id, attempt_number=self.attempt_number)



import logging
from abc import abstractmethod

from pydantic import BaseModel

from rooms_shared_services.src.lambda_handlers.abstract.handlers import AbstractNestedLambdaEventHandler
from rooms_shared_services.src.models.lmbda.dynamdb_stream import DynamodbData, DynamodbStreamEventModel

logger = logging.getLogger()
logger.setLevel(logging.INFO)


class DynamodbStreamEventHandler(AbstractNestedLambdaEventHandler):
    def provide_data_model_array(self, event_model: DynamodbStreamEventModel) -> list[BaseModel | None]:  # type: ignore
        return [self.retrieve_event_data_model(dynamodb_data=record.dynamodb) for record in event_model.Records]

    @abstractmethod
    def retrieve_event_data_model(self, dynamodb_data: DynamodbData) -> BaseModel | None:
        ...



import json
import logging
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Any, Literal, Self
from uuid import UUID, uuid4

from pydantic import BaseModel, ConfigDict

from rooms_shared_services.src.encoders.json import RawDynamodbEncoder
from rooms_shared_services.src.models.base import Unset
from rooms_shared_services.src.storage.utils import (
    convert_from_supported_formats,
    dump_value_dict,
    parse_serialized_dict,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

UNSET = Literal["UNSET"]


class BaseDynamodbModel(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def create_with_id(cls, **attributes):
        model_id = uuid4()
        value_dict = {"id": model_id, **attributes}
        return cls.model_validate(value_dict)

    def convert_value(self, item_value, exclude_unset: bool):
        match item_value:
            case BaseDynamodbModel():
                converted_item_value = item_value.dynamodb_dump(exclude_unset=exclude_unset)
            case int():
                converted_item_value = Decimal(item_value)
            case float():
                converted_item_value = Decimal(str(item_value))
            case dict():
                converted_item_value = self.convert_dict(item_value, exclude_unset=exclude_unset)
            case list():
                converted_item_value = [
                    self.convert_value(item_value_item, exclude_unset=exclude_unset) for item_value_item in item_value
                ]
            case Enum():
                converted_item_value = item_value.name
            case datetime():
                converted_item_value = item_value.isoformat()
            case _:
                converted_item_value = str(item_value)
        return converted_item_value

    def convert_to_raw(self, data_dict: dict, root_level: bool = False) -> dict:
        return {
            item_key: self.convert_to_raw_value(item_value, root_level=root_level)
            for (item_key, item_value) in data_dict.items()
        }

    def convert_to_raw_value(self, item_value: Any, root_level: bool = False):
        match item_value:
            case Decimal():
                item_value_dict = {"N": str(item_value)}
            case dict():
                item_value_dict = {"M": json.dumps(item_value, cls=RawDynamodbEncoder)}
            case list():
                item_value_dict = {"L": json.dumps(item_value, cls=RawDynamodbEncoder)}
            case str():
                item_value_dict = {"S": item_value}
            case bool():
                item_value_dict = {"BOOL": "true" if item_value else "false"}
            case None:
                item_value_dict = {"NULL": ""}
            case UUID():
                item_value_dict = {"S": str(item_value)}
            case _:
                raise ValueError("Invalid item value type")
        return item_value_dict if root_level else list(item_value_dict.values()).pop()

    def convert_dict(self, item_dict: dict, exclude_unset: bool):
        return {
            item_key: self.convert_value(item_value, exclude_unset=exclude_unset)
            for (item_key, item_value) in item_dict.items()
        }

    def dynamodb_dump(self, exclude_unset: bool = True, include: set[str] | None = None, raw: bool = False):
        item_dict = self.model_dump(include=include)
        if exclude_unset:
            item_dict = {
                item_key: item_value
                for item_key, item_value in item_dict.items()
                if not any((isinstance(item_value, Unset), item_value == "UNSET"))
            }
        return dump_value_dict(data_dict=item_dict, dynamodb_format=raw, json_format=False)

    @classmethod
    def validate_value(cls, item_value: Any) -> Any:
        match item_value:
            case list():
                validated_value = [cls.validate_value(item_value=product_elem) for product_elem in item_value]
            case "None":
                validated_value = None
            case dict():
                return {
                    elem_key: cls.validate_value(item_value=elem_value) for elem_key, elem_value in item_value.items()
                }
            case _:
                validated_value = item_value
        return validated_value

    @classmethod
    def parse_raw_record_attribute(cls, attr_dict: dict[str, str]) -> Any:
        item_key, item_value = list(attr_dict.items()).pop()
        match item_key:
            case "S":
                res = str(item_value)
            case "N":
                res = float(item_value)
            case "B":
                res = item_value
            case "BOOL":
                res = item_value == "true"
            case "NULL":
                res = None
            case "M" | "L":
                res = json.loads(item_value)
            case "SS" | "NS" | "BS":
                res = set(item_value)
            case _:
                raise ValueError("Invalid data type")
        return res

    @classmethod
    def validate_dynamodb_item(cls, data_dict: dict, from_raw: bool = False) -> Self:
        if from_raw:
            data_dict = parse_serialized_dict(value_dict=data_dict)
        data_dict = {item_key: cls.validate_value(item_value) for item_key, item_value in data_dict.items()}
        return cls.model_validate(data_dict)

    @classmethod
    def validate_stored_item_with_none(cls, data_dict: dict, from_raw: bool = False):
        value_dict = {key: data_dict.get(key) for key in cls.model_fields}
        return cls.validate_dynamodb_item(data_dict=value_dict, from_raw=from_raw)

    @classmethod
    def validate_item(cls, item_dict: dict, include_unsets: bool = False, include_nones: bool = True) -> Self:
        """Create new data model from provided data dict.

        Substitute absent values with Unset instances if include_unsets is True.
        Substitute absent values with None values if include_nones is True.
        At most one of (include_unsets, include_nones) must be True.

        Args:
            item_dict (dict): _description_
            include_unsets (bool): _description_. Defaults to False.
            include_nones (bool): _description_. Defaults to True.

        Raises:
            ValueError: when both include_unsets and include_nones are True.

        Returns:
            Self: An instance of the class.
        """
        if include_unsets + include_nones == 2:
            raise ValueError("At most one of (include_unsets, include_nones) must be True")
        model_fields = cls.model_fields
        field_names = model_fields.keys()
        key_value_dict = {}
        if include_unsets:
            key_value_dict.update({field_name: Unset() for field_name in field_names})
        if include_nones:
            key_value_dict.update({field_name: None for field_name in field_names})
        key_value_dict.update(item_dict)
        key_value_dict = {
            item_key: convert_from_supported_formats(item_value) for item_key, item_value in key_value_dict.items()
        }
        return cls.model_validate(key_value_dict)

    @classmethod
    def validate_from_raw_record(cls, data_dict: dict):
        key_value_dict = {}
        for product_key, product_value in data_dict.items():
            validated_key = product_key
            match product_key:
                case _:
                    pass
            validated_value = cls.validate_value(item_value=product_value)
            key_value_dict.update({validated_key: validated_value})

        return cls.model_validate(key_value_dict)



from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="product_attribute_source_")

    tablename: str
    region_name: str
    name_language_indexname: str



import json
from datetime import datetime
from typing import Literal

from pydantic import BaseModel, IPvAnyAddress

HttpMethod = Literal["GET", "POST", "PATCH", "DELETE"]

APIGatewayStage = Literal["stage", "prod"]


class HttpRequestEventData(BaseModel):
    method: HttpMethod
    api_id: str
    source_ip: IPvAnyAddress
    stage: APIGatewayStage
    created_at: datetime
    query_string_parameters: dict | None
    path_parameters: dict | None
    payload: dict | None

    @classmethod
    def from_lambda_event(cls, event: dict):
        model_params = {
            "method": event["httpMethod"],
            "api_id": event["requestContext"]["apiId"],
            "source_ip": event["requestContext"]["identity"]["sourceIp"],
            "stage": event["requestContext"]["stage"],
            "created_at": datetime.strptime(event["requestContext"]["requestTime"], "%d/%b/%Y:%H:%M:%S %z"),
            "query_string_parameters": event["queryStringParameters"],
            "path_parameters": event["pathParameters"],
            "payload": event["body"] and json.loads(event["body"]),
        }
        return cls.model_validate(model_params)


class CognitoIdentityData(BaseModel):
    cognito_identity_id: str | None = None
    cognito_identity_pool_id: str | None = None


class HttpRequestContextData(BaseModel):
    function_name: str
    memory_limit_in_mb: int
    function_version: str
    invoked_function_arn: str
    client_context: dict | None
    identity: CognitoIdentityData | None = None



from rooms_shared_services.src.lambda_handlers.abstract.processors import AbstractLambdaDataProcessor
from rooms_shared_services.src.storage.models import BaseDynamodbModel


class BaseDynamodbStreamEventProcessor(AbstractLambdaDataProcessor):
    model_class: BaseDynamodbModel

    def __init__(self):
        """Set items."""
        self.new_item: BaseDynamodbModel | None = None
        self.old_item: BaseDynamodbModel | None = None

    def __call__(
        self,
        keys: dict,
        new_image: dict | None = None,
        old_image: dict | None = None,
        **kwargs,
    ):  # noqa: N803
        for data_dict, attr in zip([new_image, old_image], ["new_item", "old_item"]):
            if data_dict:
                setattr(self, attr, self.validate_model(model_data=data_dict, keys=keys))
        res = super().__call__()
        self.reset()
        return res

    def reset(self):
        self.new_item = None
        self.old_item = None

    def validate_model(self, model_data: dict, keys: dict):
        model_data.update(keys)
        return self.model_class.validate_dynamodb_item(data_dict=model_data, from_raw=True)



import logging
from decimal import Decimal
from typing import Any, Generator, Literal

import boto3
from boto3.dynamodb.conditions import Attr, Key
from more_itertools import batched
from mypy_boto3_dynamodb import DynamoDBClient
from mypy_boto3_dynamodb.service_resource import DynamoDBServiceResource, Table
from pydantic_settings import BaseSettings

from rooms_shared_services.src.exceptions.storage import MoreThanOneItemFound
from rooms_shared_services.src.storage.abstract import AbstractStorageClient
from rooms_shared_services.src.storage.models import BaseDynamodbModel

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

ConditionLiteral = Literal["OR", "AND"]
FilterAttrCondition = Literal[
    "attribute_type",
    "begins_with",
    "between",
    "contains",
    "eq",
    "exists",
    "gt",
    "gte",
    "is_in",
    "lt",
    "lte",
    "ne",
    "not_exists",
    "size",
]


class DynamodbSettings(BaseSettings):
    table: str


class DynamodbStorageClient(AbstractStorageClient):
    def __init__(
        self,
        tablename: str,
        region_name: str = "us-east-1",
        endpoint_url: str | None = None,
        create_table_params: dict | None = None,
    ):
        """Set attributes.Create db table if needed.

        Args:
            tablename (str): _description_
            region_name (str): _description_. Defaults to "us-east-1".
            endpoint_url (str | None): _description_. Defaults to None.
            create_table_params (dict | None): _description_. Defaults to None.
        """
        resource_params = {
            "region_name": region_name,
        }
        if endpoint_url:
            resource_params.update({"endpoint_url": endpoint_url})
        dynamodb_resource: DynamoDBServiceResource = boto3.resource("dynamodb", **resource_params)  # type: ignore
        if create_table_params:
            if "ProvisionedThroughput" not in create_table_params:
                create_table_params.update(
                    {
                        "ProvisionedThroughput": {
                            "ReadCapacityUnits": 5,
                            "WriteCapacityUnits": 5,
                        },
                    },
                )
            dynamodb_resource.create_table(TableName=tablename, **create_table_params)
        self.table: Table = dynamodb_resource.Table(tablename)  # type: ignore
        self.client: DynamoDBClient = boto3.client("dynamodb", **resource_params)
        logger.info("Dynamodb storage slient initiated with table {}".format(self.table.name))

    def __call__(self):
        logger.info("Hello world")

    def retrieve(self, key: dict, **call_params) -> dict | None:
        response = self.table.get_item(Key=key, **call_params)
        try:
            return response["Item"]
        except KeyError:
            return None

    def get_data_model(self, key: dict, model_type: type[BaseDynamodbModel], **call_params):
        data_item = self.retrieve(key=key, **call_params)
        if data_item is None:
            return None
        return model_type.validate_stored_item_with_none(data_item)

    def index_query(  # noqa: WPS211
        self,
        index_name: str,
        partition_key_name: str,
        partition_key_value: str,
        sort_key_name: str | None = None,
        sort_key_value: str | None = None,
        **call_params,
    ) -> dict:
        key_condition_expression = Key(partition_key_name).eq(partition_key_value)
        if sort_key_name and sort_key_value:
            key_condition_expression = key_condition_expression & Key(sort_key_name).eq(sort_key_value)
        return self.table.query(IndexName=index_name, KeyConditionExpression=key_condition_expression)

    def retrieve_from_index(self, index_name: str, partition_key_name: str, partition_key_value: str, **call_params):
        resp = self.index_query(
            index_name=index_name,
            partition_key_name=partition_key_name,
            partition_key_value=partition_key_value,
            **call_params,
        )
        match resp["Count"]:
            case 1:
                return resp["Items"][0]
            case 0:
                return None
            case _:
                raise MoreThanOneItemFound(resp["Items"])

    def list_from_index(self, index_name: str, partition_key_name: str, partition_key_value: str, **call_params):
        resp = self.index_query(
            index_name=index_name,
            partition_key_name=partition_key_name,
            partition_key_value=partition_key_value,
            **call_params,
        )
        match resp["Count"]:
            case 0:
                return []
            case _:
                return resp["Items"]

    def create(self, table_item: dict, **call_params) -> dict:
        return dict(self.table.put_item(Item=table_item, **call_params))

    def update(self, key: dict, attribute_updates: dict, **call_params) -> dict:
        update_expression = "SET "
        expression_attribute_values = {}
        attribute_update_items = attribute_updates.items()
        for idx, update in enumerate(attribute_update_items):
            value_name = ":val{}".format(idx)
            update_value = update[1]
            update_value = Decimal(str(update_value)) if isinstance(update_value, (int, float)) else update_value
            update_expression = "{} {} = {},".format(update_expression, update[0], value_name)
            expression_attribute_values[value_name] = update_value
        update_expression = update_expression[:-1]
        print(
            "update item: key {}, UpadteExspression {}, ExpressionAttributeValues {}".format(
                key,
                update_expression,
                expression_attribute_values,
            ),
        )
        return dict(
            self.table.update_item(
                Key=key,
                UpdateExpression=update_expression,
                ExpressionAttributeValues=expression_attribute_values,
                **call_params,
            ),
        )

    def delete(self, key: dict, **call_params) -> dict:
        return dict(self.table.delete_item(Key=key, **call_params))

    def bulk_retrieve(self, keys: list[dict], **call_params) -> list[dict | None]:
        return [self.retrieve(key=key, **call_params) for key in keys]

    def bulk_create(self, table_items: list[dict], **call_params) -> None:
        with self.table.batch_writer() as batch:
            for table_item in table_items:
                batch.put_item(Item=table_item, **call_params)

    def bulk_update(self, keys: list[dict[Any, Any]], attribute_updates_list: list[dict], **call_params) -> list[dict]:
        if len(keys) == len(attribute_updates_list):
            responses = []
            for key, attribute_updates in zip(keys, attribute_updates_list):
                responses.append(self.update(key=key, attribute_updates=attribute_updates, **call_params))
            return responses
        raise ValueError("Keys and attribute_updates_list must be of equal size")

    def bulk_delete(self, keys: list[dict], **call_params) -> None:
        with self.table.batch_writer() as batch:
            for key in keys:
                batch.delete_item(Key=key, **call_params)

    def bulk_get(
        self,
        filter_params: dict | None = None,
        condition: ConditionLiteral = "AND",
        filter_attr_condition: FilterAttrCondition = "eq",
        negate: bool = False,
    ):
        scan_params = self.create_scan_params(
            filter_params=filter_params,
            condition=condition,
            filter_attr_condition=filter_attr_condition,
            negate=negate,
        )
        resp = self.table.scan(**scan_params)
        return resp["Items"]

    def create_scan_params(
        self,
        filter_attr_condition: FilterAttrCondition,
        condition: ConditionLiteral,
        negate: bool,
        filter_params: dict | None = None,
    ):
        scan_params = {}
        if filter_params:
            FilterExpression = self.create_filter_expression(  # noqa: N806
                filter_params=filter_params,
                filter_attr_condition=filter_attr_condition,
                condition=condition,
                negate=negate,
            )  # noqa: N806
            scan_params.update({"FilterExpression": FilterExpression})
        return scan_params

    def create_filter_expression(
        self,
        filter_params: dict,
        filter_attr_condition: FilterAttrCondition,
        condition: ConditionLiteral,
        negate: bool = False,
    ):
        filter_params_list = list(filter_params.items())
        first_params_item = filter_params_list.pop(0)
        attr_method = getattr(Attr(first_params_item[0]), filter_attr_condition)
        FilterExpression = attr_method(first_params_item[1])  # noqa: N806
        for next_params_item in filter_params_list:
            match condition:
                case "AND":
                    FilterExpression = FilterExpression & Attr(next_params_item[0]).eq(  # type: ignore # noqa: N806
                        next_params_item[1],
                    )
                case "OR":
                    FilterExpression = FilterExpression | Attr(next_params_item[0]).eq(  # type: ignore # noqa: N806
                        next_params_item[1],
                    )
                case _:
                    raise ValueError("Invalid filter expression condition")
        if negate:
            FilterExpression = ~(FilterExpression)  # noqa: N806
        return FilterExpression

    def validate_data_elem(self, data_elem_value: dict) -> str | int | float | dict | list | None:
        data_type = list(data_elem_value.keys())[0]
        parsed_value = data_elem_value[data_type]
        validated_value: str | int | float | dict | list | None
        match data_type:
            case "S":
                assert isinstance(parsed_value, str)
                validated_value = parsed_value
                if validated_value == "None":
                    validated_value = None
            case "N":
                assert isinstance(parsed_value, str)
                if "." in parsed_value:
                    validated_value = float(parsed_value)
                else:
                    validated_value = int(parsed_value)
            case "BOOL":
                match parsed_value:
                    case "false":
                        validated_value = False  # noqa: WPS220
                    case "true":
                        validated_value = True  # noqa: WPS220
                    case _:
                        raise ValueError("invalid boolean")  # noqa: WPS220
            case "NULL":
                validated_value = None
            case "M":
                assert isinstance(parsed_value, dict)
                validated_value = {
                    elem_key: self.validate_data_elem(elem_value) for elem_key, elem_value in parsed_value.items()
                }
            case "L":
                validated_value = [self.validate_data_elem(elem_item) for elem_item in parsed_value]
        return validated_value

    def parse_annotated_response(self, data_item: dict):
        return {
            data_elem_key: self.validate_data_elem(data_elem_value)
            for data_elem_key, data_elem_value in data_item.items()
        }

    @staticmethod
    def create_value_type(attr_value):
        match attr_value:
            case int():
                return {"N": str(attr_value)}
            case str():
                return {"S": attr_value}

    def create_pagination_filter_expression(self, filter_params: dict):
        filter_elements = [
            ("{} = :name_{}".format(key, idx), (":name_{}".format(idx), self.create_value_type(key_value)))
            for idx, (key, key_value) in enumerate(filter_params.items())
        ]
        expression_elements = [elem[0] for elem in filter_elements]
        attribute_elements = [elem[1] for elem in filter_elements]
        filter_expression = ",".join(expression_elements)
        attribute_value_dict = dict(attribute_elements)
        return filter_expression, attribute_value_dict

    def list_for_partition_key(
        self,
        partition_key_name: str,
        partition_key_value: str,
        item_model_type: type[BaseDynamodbModel] | None = None,
        ascending: bool = True,
    ):
        key_condition_expression = Key(partition_key_name).eq(partition_key_value)
        resp = self.table.query(KeyConditionExpression=key_condition_expression, ScanIndexForward=ascending)
        table_items = resp["Items"]
        if item_model_type:
            return [item_model_type.validate_stored_item_with_none(data_dict=table_item) for table_item in table_items]
        return table_items

    def retrieve_for_partition_key(
        self,
        partition_key_name: str,
        partition_key_value: str,
        item_model_type: type[BaseDynamodbModel] | None = None,
        ascending: bool = True,
    ):
        query_list = self.list_for_partition_key(
            partition_key_name=partition_key_name,
            partition_key_value=partition_key_value,
            item_model_type=item_model_type,
            ascending=ascending,
        )
        return query_list[0]

    def batch_retrieve(self, keys: list[dict[Any, Any]]):
        res_list = []
        for chunk in batched(keys, 100):
            res = self.retrieve_by_chunk(keys=chunk)
            listed_res = [res[chunk_item["id"]] for chunk_item in chunk]
            res_list.extend(listed_res)
        return res_list

    def retrieve_by_chunk(
        self,
        keys: list[dict[Any, Any]] | None = None,
        table_items: list[dict] | None = None,
        request_items: dict[str, dict] | None = None,
    ):
        table_items = table_items or []
        if not request_items and keys is not None:
            key_list = [
                {key_item[0]: self.create_value_type(key_item[1]) for key_item in key_entry.items()}
                for key_entry in keys
            ]
            request_items = {self.table.name: {"Keys": key_list}}
        resp = self.client.batch_get_item(RequestItems=request_items)
        table_items.extend(resp["Responses"][self.table.name])
        if resp["UnprocessedKeys"]:
            return self.retrieve_by_chunk(table_items=table_items, request_items=resp["UnprocessedKeys"])
        parsed_response = [self.parse_annotated_response(data_item=table_item) for table_item in table_items]
        return {resp_item["id"]: resp_item for resp_item in parsed_response}

    def batch_update(self, keys: list[dict[Any, Any]], attribute_updates_list: list[dict], **call_params):
        if None in attribute_updates_list:
            raise ValueError("Null attribute update item is not allowed")
        stored_items = self.batch_retrieve(keys=keys)
        if None in stored_items:
            raise ValueError("Could not retrieve a table item for update")
        remove_keys = [
            [remove_key for remove_key in attribute_update_dict.keys() if attribute_update_dict[remove_key] is None]
            for attribute_update_dict in attribute_updates_list
        ]
        updated_items = [{**table_item, **update} for table_item, update in zip(stored_items, attribute_updates_list)]
        updated_items = [
            {item_key: item_value for item_key, item_value in updated_items.items() if item_key not in remove_key_elems}
            for updated_items, remove_key_elems in zip(updated_items, remove_keys)
        ]
        updated_items = [self.validate_item_values(updated_item) for updated_item in updated_items]
        with self.table.batch_writer() as batch:
            for updated_item in updated_items:
                batch.put_item(Item=updated_item)
        logger.info("Batch update completed at table {}".format(self.table.name))

    def validate_item_values(self, table_item_dict: dict):
        return {key: self.convert_value(item_value) for key, item_value in table_item_dict.items()}

    def convert_value(self, table_item_value: Any):
        match table_item_value:
            case float():
                return Decimal(str(table_item_value))
            case dict():
                return {key: self.convert_value(item_value) for key, item_value in table_item_value.items()}
            case _:
                return table_item_value

    def iterate_annotated_responses(self, table_items):
        for db_item in table_items:  # noqa: WPS526
            yield self.parse_annotated_response(db_item)

    def get_by_pages(
        self,
        batch_size: int | None = None,
        page_size: int = 100,
        consistent_read: bool = False,
        filter_by: dict | None = None,
    ):
        batch_size = batch_size or page_size
        yield from batched(
            self.iterate_table_items(
                page_size=page_size,
                consistent_read=consistent_read,
                filter_by=filter_by,
            ),
            batch_size,
        )

    def iterate_table_item_models(self, item_model_type: BaseDynamodbModel, table_items: list[dict]):
        for annotated_table_item in self.iterate_annotated_responses(table_items=table_items):  # noqa: WPS526
            yield item_model_type.validate_dynamodb_item(annotated_table_item)

    def clean_field(self, fieldname: str, key_fields: list[str] | None = None):
        key_fields = key_fields or ["id"]
        counter = 0
        for page in self.get_by_pages():
            keys = [{key_field: table_item[key_field] for key_field in key_fields} for table_item in page]
            attribute_updates_list = [{fieldname: None} for _ in keys]
            logger.info("Attribute update list creaed for cleaning.")
            self.batch_update(keys=keys, attribute_updates_list=attribute_updates_list)
            counter += len(page)
            logger.info("{} elements cleaned".format(counter))

    def iterate_table_items(
        self,
        page_size: int | None = 100,
        consistent_read: bool = False,
        filter_by: dict | None = None,
        item_model_type: BaseDynamodbModel | None = None,
    ) -> Generator[list[dict], None, None]:
        paginator = self.client.get_paginator("scan")
        scan_params = {"TableName": self.table.name}
        if filter_by:
            filter_expression, attribute_values = self.create_pagination_filter_expression(filter_params=filter_by)
            scan_params.update({"FilterExpression": filter_expression, "ExpressionAttributeValues": attribute_values})
        if page_size:
            scan_params.update({"PaginationConfig": {"PageSize": page_size}})
        for page in paginator.paginate(**scan_params, ConsistentRead=consistent_read):
            table_items = page["Items"]
            logger.info("Collected {} page items from dynamodb table {}".format(len(table_items), self.table.name))
            if item_model_type:
                yield from self.iterate_table_item_models(table_items=table_items, item_model_type=item_model_type)
            else:
                yield from self.iterate_annotated_responses(table_items=table_items)



import json
from abc import ABC, abstractmethod

from openai import OpenAI
from pydantic import BaseModel

from rooms_shared_services.src.models.texts.variants import LLMRequestVariant


class OpenaiRequestMessage(BaseModel):
    role: str
    content: str  # noqa: WPS110


class AbstractLLMJSONQueryClient(ABC):
    ...


class AbstractOpenaiJSONQueryClient(AbstractLLMJSONQueryClient):
    def __init__(
        self,
        request_variant: LLMRequestVariant,
        openai_model: str = "gpt-3.5-turbo",
        retry_count: int = 25,
    ):
        """Set attributes.

        Args:
            openai_model (str): __description__.
            request_variant (LLMRequestVariant): _description_.
            retry_count (int): _description_. Defaults to 25
        """
        self.openai_model = openai_model
        self.request_variant = request_variant
        self.openai_client = OpenAI()
        self.retry_count = retry_count

    def run_query(self, **request_params) -> str | dict:
        for _ in range(self.retry_count):
            messages = self.collect_messages(request_variant=self.request_variant, **request_params)
            response = self.receive_response(messages=messages)
            response = response.choices[0].message.content
            validated_response = self.validate_json_response(response=response, **request_params)
            if validated_response:
                return validated_response
        raise ValueError("No valid response received")

    @abstractmethod
    def collect_messages(self, **request_params) -> list[OpenaiRequestMessage]:
        ...

    def receive_response(self, messages: list[OpenaiRequestMessage]):
        return self.openai_client.chat.completions.create(
            model=self.openai_model,
            messages=[message.model_dump() for message in messages],
        )

    def validate_json_response(self, response: str, **kwargs) -> str | dict | None:
        print(response)
        try:
            return json.loads(response)["result"]
        except Exception:
            return None
